# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum Dresden - Rossendorf (HZDR)
#
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional, Tuple

import ipywidgets as widgets
import numpy as np
import numpy.typing as npt
import pandas as pd
from IPython.display import clear_output, display
from ipywidgets import (
    HTML,
    AppLayout,
    GridspecLayout,
    HBox,
    Image,
    Label,
    Output,
    VBox,
)
from scipy import integrate

from metabolator import files, fit, plots


def _example_datasets_html():
    datasets = files.get_example_dataset_urls()

    if not datasets:
        return ""

    downloads = (
        f'<code><a download="{name}" href="{url}">{name}</a></code>'
        for name, url in datasets.items()
    )
    return f"<p>Example datasets: {', '.join(downloads)}</p>"


class ThermalAnalysisApp:
    #: Format string used to build download links
    DOWNLOAD_LINK_HTML = '<a href="{href}" download="{download}">{text}</a>'

    HEADER_HTML = """
        <h1>METABOLATOR</h1>
        <h2>Analysis of Microcalorimetric Metabolic Data Using Monod's Equation</h2>
        <p>If you use METABOLATOR for your data analysis, please cite the software
          <a target="_blank"
             href="{software_doi_url}"
             title="Cite the software!"
             rel="noreferrer">
            <img alt="DOI {software_doi}"
                 src="{software_badge_url}"
                 referrerpolicy="no-referrer" />
          </a>
          and this service
          <a target="_blank"
             href="{service_doi_url}"
             title="Cite this service!"
             rel="noreferrer">
            <img alt="DOI {service_doi}"
                 src="{service_badge_url}"
                 referrerpolicy="no-referrer" />
          </a>
        </p>
        {example_datasets_html}
    """.format(
        software_doi_url="https://doi.org/10.14278/rodare.3050",
        software_doi="10.14278/rodare.3050",
        software_badge_url="https://rodare.hzdr.de/badge/DOI/10.14278/rodare.3050.svg",
        service_doi_url="https://doi.datacite.org/dois/10.14278%2F1e4y-rc46",
        service_doi="10.14278/1e4y-rc46",
        service_badge_url="https://img.shields.io/badge/DOI-10.14278%2F1e4y--rc46-blue",
        example_datasets_html=_example_datasets_html(),
    )

    FOOTER_HTML = """
        Helmholtz-Zentrum Dresden - Rossendorf
        <ul>
          <li><a href="{imprint_url}">Imprint</a></li>
          <li><a href="{privacy_url}">Privacy</a></li>
          <li>Contact: {contact_address}</li>
        </ul>
    """.format(
        imprint_url="https://www.hzdr.de/impressum",
        privacy_url="/privacy.html",
        contact_address="metabolator@hzdr.de",
    )

    HEAT_RANGE_DEFAULT_VALUES = (0.34, 0.87)
    SLIDER_STEP_SIZE = 0.01
    SI_TO_MICRO_FACTOR = 1e6
    MICRO_TO_SI_FACTOR = 1e-6

    # Percentage variation from peak in either direction where the fitting procedure is
    # expected to work well
    ESTIMATE_UPPER_RANGE = 0.15
    ESTIMATE_LOWER_RANGE = 0.60

    # Fit size
    FIT_SIZE = 500

    def __init__(self):
        # Initialize widgets
        self.upload = widgets.FileUpload(
            accept=".xls,.xlsx,.xlsm,.xlsb,.odf,.ods,.odt",
            description="Browse files",
        )
        self.upload_button = widgets.Button(description="Start Upload")
        self.time_offset = widgets.FloatText(description="Time Offset (s):")
        self.power_offset = widgets.FloatText(description="Power Offset (µW):")
        self.all_opt = []
        self.dataset_name: str = None
        self.selector = widgets.Select(description="Select the samples")

        self.analysis_results = pd.DataFrame()

        self.heat_range_slider = widgets.FloatRangeSlider(
            value=list(self.HEAT_RANGE_DEFAULT_VALUES),
            min=0,
            max=1,
            step=self.SLIDER_STEP_SIZE,
            description="Heat Range (J):",
            readout_format=".2f",
        )
        self.estimated_heat_range_button = widgets.Button(
            description="Estimate Heat Range"
        )
        self.check_offsetting_button = widgets.Button(description="Check Offsetting")

        self.plot_output = Output()
        self.log_output = Output()

        self.enthalpy_plot_button = widgets.Button(description="Enthalpy Plot")
        self.time_plot_button = widgets.Button(description="Time Plot")
        self.analyze_data_button = widgets.Button(description="Analyze Data / Fit")

        # Bind button events
        self.enthalpy_plot_button.on_click(self.enthalpy_plot)
        self.time_plot_button.on_click(self.time_plot)
        self.check_offsetting_button.on_click(self.offset_plot)
        self.estimated_heat_range_button.on_click(self.estimated_heat_range)
        self.analyze_data_button.on_click(self.analyze_data)

        self.upload_button.on_click(self.on_upload_button_clicked)

        self.dataset_selection = VBox(
            [
                Label("1. Dataset selection", style={"font_weight": "bold"}),
                HBox([self.upload, self.upload_button]),
                self.selector,
            ],
        )

        self.ranges = VBox(
            [
                Label("2. Select Ranges", style={"font_weight": "bold"}),
                VBox(
                    [
                        self.time_offset,
                        self.power_offset,
                        self.heat_range_slider,
                        HBox(
                            [
                                self.estimated_heat_range_button,
                                self.check_offsetting_button,
                            ]
                        ),
                    ],
                ),
            ],
        )

        self.data_link = widgets.HTML(
            value=self.DOWNLOAD_LINK_HTML.format(href="", download="", text="-")
        )

        self.downloads = HBox(
            [
                Label(value="Download:", style={"font_weight": "bold"}),
                self.data_link,
            ],
        )

        self.create_plots = VBox(
            [
                Label(value="3. Plot and Analyze", style={"font_weight": "bold"}),
                VBox(
                    [
                        self.enthalpy_plot_button,
                        self.time_plot_button,
                        self.analyze_data_button,
                    ]
                ),
                self.downloads,
            ],
        )

        self.center_grid = GridspecLayout(6, 3)

        self.center_grid[0, 0] = self.dataset_selection
        self.center_grid[0, 1] = self.ranges
        self.center_grid[0, 2] = self.create_plots

        self.center_grid[1:, :2] = self.plot_output
        self.center_grid[1:, 2] = self.log_output

        self.layout = AppLayout(
            header=HTML(value=self.HEADER_HTML),
            left_sidebar=None,
            center=self.center_grid,
            right_sidebar=None,
            footer=HBox(
                [
                    Image(value=files.get_resource_contents("hzdr-logo.png")),
                    HTML(value=self.FOOTER_HTML),
                ]
            ),
        )

        self.action_buttons: List[widgets.Button] = [
            self.estimated_heat_range_button,
            self.enthalpy_plot_button,
            self.time_plot_button,
            self.check_offsetting_button,
            self.analyze_data_button,
        ]

        self.disable_buttons()

        # This holds the data we're working on
        self.df: Optional[pd.DataFrame] = None

    def run(self):
        """Display the app."""
        display(self.layout)

    def load_data(self, filename: Optional[str] = None):
        """Load filename and data into the app.

        If ``filename`` is provided, the file with this name is loaded from the package
        resources. Otherwise, data and filename from the file upload widget are used.

        Stored results from previous analysis runs are reset.
        """
        self.analysis_results = pd.DataFrame()
        self.data_link.value = self.DOWNLOAD_LINK_HTML.format(
            href="", download="", text="-"
        )

        content: Optional[bytes] = None
        if filename is None:
            uploaded_file = self.upload.value[0]
            filename = uploaded_file["name"]
            buffer: memoryview = uploaded_file["content"]
            content = buffer.tobytes()

        try:
            self.dataset_name, self.df = files.load_excel_file_data(filename, content)
        except files.MalformedDataFileError:
            self.dataset_name = None
            self.df = None

        self.enable_buttons()

    def on_upload_button_clicked(self, _button, dataset: Optional[str] = None):
        self.load_data(filename=dataset)
        if self.dataset_name is None or self.df is None:
            with self.log_output:
                clear_output(wait=True)
                print("Failed to load dataset")
        else:
            self.update_options()
            self.update_range_sliders()
            with self.log_output:
                clear_output(wait=True)
                print("Loaded dataset", self.dataset_name)

    def update_options(self):
        """Update the measurement selection box from ``self.df`` contents."""
        self.all_opt = list(self.df.columns)
        self.selector.options = list(self.df.columns)

    def update_range_sliders(self):
        """Update the heat range slider maximum value based on the loaded data."""
        max_heat = []
        time = self.df.index.to_numpy()
        for i in range(len(self.all_opt)):
            power = self.df.iloc[:, i].to_numpy()
            # Calculate enthalpy in Joules
            th = integrate.cumulative_trapezoid(power, time)
            max_heat.append(max(th))
        self.heat_range_slider.max = max(max_heat)

    def disable_buttons(self):
        """Disable all action buttons."""
        for button in self.action_buttons:
            button.disabled = True

    def enable_buttons(self):
        """Enable all action buttons."""
        for button in self.action_buttons:
            button.disabled = False

    def prepare_data(
        self,
        measurement_index: int,
        *,
        power_offset: float | None = None,
        time_offset: float | None = None,
        lower_power_limit: bool = False,
    ) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
        """Get measurement from dataset, calculate enthalpy, and apply offsetting.

        ``time_offset`` is applied by finding the time step closest (in any direction)
        to the desired offset and discarding all values before it. If no time offset is
        given, no values are discarded in this step.

        ``power_offset`` is applied by subtracting it from all power values. If no power
        offset is given, an offset of `0.0` is assumed.

        Finally, if ``lower_power_limit`` is set, ``self.curve_start()`` is used to
        determine the beginning of the curve. All values before it, are discarded.

        Values are always discarded in all arrays (power, time, enthalpy) to keep them
        at the same length.

        Power, time, and enthalpy are returned as separate arrays.
        """

        # Conversion to µW
        data = self.df.iloc[:, measurement_index] * self.SI_TO_MICRO_FACTOR

        power = data.to_numpy()
        time = data.index.to_numpy()

        start = 0
        if time_offset is not None:
            # Find time point closest to desired time offset
            distance_to_offset = np.abs(time - time_offset)
            start = np.nanargmin(distance_to_offset)

        # default power offset is 0.0
        power_offset = power_offset or 0.0

        # Apply offestting in all directions
        power = power[start:] - power_offset
        time = time[start:] - time[start]

        # Calculate enthalpy in Joules after applying offsets.
        # An initial value has to be given in order to keep to same length as power
        # and time.
        th = (
            integrate.cumulative_trapezoid(power, time, initial=0)
            * self.MICRO_TO_SI_FACTOR
        )

        # Exclude negative power values
        if lower_power_limit:
            start = self.curve_start(power)
            th = th[start:]
            power = power[start:]
            time = time[start:]

        return power, time, th

    def estimated_heat_range(self, _button: widgets.Button):
        """Estimate a heat range for fitting.

        This uses empirical values to calculates a range which is likely to deliver good
        fitting results.

        The range is calculated an sliders are set accordingly.
        """
        selection = self.selector.index or 0

        power, time, th = self.prepare_data(
            selection,
            power_offset=self.power_offset.value,
            time_offset=self.time_offset.value,
        )

        max_index = np.argmax(power)
        peak_x = th[max_index]

        low = (1.0 - self.ESTIMATE_LOWER_RANGE) * peak_x
        high = (1.0 + self.ESTIMATE_UPPER_RANGE) * peak_x
        self.heat_range_slider.value = [low, high]

    def enthalpy_plot(self, _button: widgets.Button):
        """Show enthalpy plot for the selected measurement."""
        selection = self.selector.index or 0
        power, time, th = self.prepare_data(selection, lower_power_limit=True)

        figure = plots.enthalpy(th, power, self.all_opt[selection])
        with self.plot_output:
            clear_output(wait=True)
            figure.show()

    def time_plot(self, _button: widgets.Button):
        """Show time plot for the selected measurement."""
        selection = self.selector.index or 0
        power, time, th = self.prepare_data(selection)

        figure = plots.time(time, power, self.all_opt[selection])
        with self.plot_output:
            clear_output(wait=True)
            figure.show()

    def offset_plot(self, _button: widgets.Button):
        """Show time plot for the selected measurement with offsetting applied."""
        selection = self.selector.index or 0
        power, time, th = self.prepare_data(
            selection,
            power_offset=self.power_offset.value,
            time_offset=self.time_offset.value,
            lower_power_limit=True,
        )

        figure = plots.offsetted(time, power, self.all_opt[selection])
        with self.plot_output:
            clear_output(wait=True)
            figure.show()

    def update_analysis_results(self, new_results: pd.DataFrame):
        self.analysis_results = pd.concat([self.analysis_results, new_results], axis=1)

        means = self.analysis_results.loc[:, (slice(None), "Values")].mean(axis=1)
        if "Means" not in self.analysis_results.columns:
            self.analysis_results.insert(0, "Means", means)
        else:
            self.analysis_results["Means"] = means

        analysis_filename = f"{self.dataset_name}_analysis.xlsx"
        self.data_link.value = self.DOWNLOAD_LINK_HTML.format(
            href=files.make_excel_file_data_url(self.analysis_results),
            download=analysis_filename,
            text=analysis_filename,
        )

    def analyze_data(self, button: widgets.Button):
        """Run the analysis on the selected measurement and show fitting plots."""

        button_description = button.description
        button.description = "Analysis in progress ..."
        self.disable_buttons()

        selection = self.selector.index or 0
        power, time, th = self.prepare_data(
            selection,
            power_offset=self.power_offset.value,
            time_offset=self.time_offset.value,
            lower_power_limit=True,
        )

        mask = (th >= self.heat_range_slider.value[0]) & (
            th <= self.heat_range_slider.value[1]
        )

        start_index = np.where(mask)[0][0]
        end_index = np.where(mask)[0][-1]

        xc, yc = th[start_index:end_index], power[start_index:end_index]

        llim, ulim, result, r_squared = fit.do_fit(xc, yc)

        rz = result.params["rz"]
        hz = result.params["hz"]
        hf = result.params["hf"]
        delta = result.params["delta"]
        Pmax = np.max(yc)  # noqa: N806
        Hmax = xc[np.argmax(yc)]  # noqa: N806

        x = np.linspace(0 - delta, hz - delta, self.FIT_SIZE)
        y = fit.model(x, rz, hz, hf, delta)
        y = y[y >= 0]

        x_fit = np.linspace(min(xc), max(xc), self.FIT_SIZE)
        y_fit = fit.model(x_fit, rz, hz, hf, delta)

        measurement_label = self.all_opt[selection]
        index = pd.Index(
            ["llim", "ulim", "r_sq", "rz", "delta", "hz", "hf", "Pmax", "Hmax"]
        )
        columns = pd.MultiIndex.from_tuples(
            [(measurement_label, "Values"), (measurement_label, "Standard Errors")]
        )

        report_data = pd.DataFrame(
            [
                [self.heat_range_slider.value[0], None],
                [self.heat_range_slider.value[1], None],
                [r_squared, None],
                [rz.value, rz.stderr],
                [delta.value, delta.stderr],
                [hz.value, hz.stderr],
                [hf.value, hf.stderr],
                [Pmax, None],
                [Hmax, None],
            ],
            index=index,
            columns=columns,
        )

        figure = plots.plot_fitting(
            (x, xc, th, x_fit),
            (y, yc, power, y_fit),
            measurement_label,
        )

        with self.plot_output:
            clear_output(wait=True)
            figure.show()

        with self.log_output:
            clear_output(wait=True)
            print(
                self.dataset_name,
                measurement_label,
                f"Heat range selection: {list(self.heat_range_slider.value)}",
                f"Fitted interval: [{llim:.8f}, {ulim:.8f}]",
                "",
                f"r_sq:  {r_squared:13.8f}",
                f"rz:    {rz.value:13.8f} +/- {rz.stderr:10.8f}",
                f"delta: {delta.value:13.8f} +/- {delta.stderr:10.8f}",
                f"hz:    {hz.value:13.8f} +/- {hz.stderr:10.8f}",
                f"hf:    {hf.value:13.8f} +/- {hf.stderr:10.8f}",
                f"Pmax:  {Pmax:13.8f}",
                f"Hmax:  {Hmax:13.8f}",
                sep="\n",
                end="\n\n",
            )

        self.update_analysis_results(report_data)

        t = np.zeros_like(x)
        dx = np.diff(x, prepend=0) * self.SI_TO_MICRO_FACTOR
        dt = dx / y
        for j in range(1, len(t)):
            t[j] = t[j - 1] + dt[j]

        max_x_index = np.argmax(y)
        max_p_index = np.argmax(power)
        t = t - t[max_x_index] + time[max_p_index]

        fig_t = plots.plot_fitting_time((t, time), (y, power))

        with self.plot_output:
            fig_t.show()

        button.description = button_description
        self.enable_buttons()

    @staticmethod
    def curve_start(arr: npt.NDArray) -> int:
        r"""Find the start of a curve.

        The starting point is assumed to be the last non-negative point encountered when
        going left from the maximum of the curve.

        In the schematic below, this would be the point marked with the asterisk.

        .. code:

                     ^
                    / \       ^
               ^   /   \     / \   ^
           ---/-\-*-----\---/---\-/-\---
                 v       \ /     v
                          v

        Return the index of the found value, or 0 if no value is found.
        """
        index = np.argmax(arr)
        while index > 0 and arr[index] >= 0:
            index = index - 1
        return index
