# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum Dresden - Rossendorf (HZDR)
#
# SPDX-License-Identifier: Apache-2.0

"""Utilities for file parsing and creation, and loading package resources."""

import base64
import re
from importlib import resources
from io import BytesIO
from pathlib import Path
from typing import Dict, Optional

import pandas as pd

MIME_TYPES = {
    ".xls": "application/vnd.ms-excel",
    ".xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
}


class UnsupportedFileTypeError(TypeError):
    def __init__(self, filename: str, extension: str):
        super().__init__(f"Unsupported file type '{extension}': {filename}")


class MalformedDataFileError(ValueError):
    def __init__(self, filename: str) -> None:
        super().__init__(f"Data file is malformed and could not be parsed: {filename}")


def data_url(mime_type: str, data: bytes) -> str:
    """Create a base64-encoded data URL for ``data``."""
    data_b64 = base64.standard_b64encode(data).decode()
    return f"data:{mime_type};base64,{data_b64}"


def get_resource_contents(filename: str) -> bytes:
    """Get a file from the static directory and return its contents as bytes."""
    static_directory = resources.files("metabolator").joinpath("static")
    return static_directory.joinpath(filename).read_bytes()


def get_resource_data_url(filename: str) -> str:
    """Get a file from the static directory and return its contents as a data URL."""
    extension = Path(filename).suffix
    if extension not in MIME_TYPES:
        raise UnsupportedFileTypeError(filename, extension)

    mime_type = MIME_TYPES[extension]
    data = get_resource_contents(filename)
    return data_url(mime_type, data)


def make_excel_file_contents(data: pd.DataFrame) -> bytes:
    """Create an Excel (xlsx) file from ``data`` and return it as bytes."""
    excel_file = BytesIO()
    with pd.ExcelWriter(excel_file, engine="xlsxwriter") as writer:
        data.to_excel(writer)
    # move cursor to beginning of the buffer
    excel_file.seek(0)
    return excel_file.read()


def make_excel_file_data_url(data: pd.DataFrame) -> str:
    """Create an Excel (xlsx) file from ``data`` and convert it into a data URL."""
    excel_file_data = make_excel_file_contents(data)
    mime_type = MIME_TYPES[".xlsx"]
    return data_url(mime_type, excel_file_data)


def load_excel_file_data(
    filename: str, content: Optional[bytes] = None
) -> tuple[str, pd.DataFrame]:
    """Load data from an Excel file.

    If ``content`` is ``None``, ``filename`` is used to load the file from the package
    resources.

    The data is read in two passes: The first pass figures out which worksheet to use
    and how many header rows exist in the sheet. The second pass uses this information
    to then interpret the data correctly.

    The name of the dataset and its contents are returned as a tuple.
    """
    if content is None:
        content = get_resource_contents(filename)

    # first pass
    # NOTE: Using bytes in read_excel is deprecated -- convert into ``BytesIO``.
    # NOTE: read_excel calls ``close`` on the buffer which discards it. Thus, in the
    #  second call to ``read_excel``, we need to create a new buffer.
    sheets = pd.read_excel(
        BytesIO(content),
        sheet_name=None,  # get all sheets
        header=None,  # read headers as normal rows
    )

    # use "Raw data" worksheet if it exists, otherwise use the first one
    sheet_name = "Raw data"
    data: Optional[pd.DataFrame] = sheets.get(sheet_name)
    if data is None:
        sheet_name = next(iter(sheets.keys()))
        data = sheets[sheet_name]

    if data.empty:
        raise MalformedDataFileError(filename)

    # In the first column, count the number of fields until time values start. This is
    # achieved by coercing non-numeric values into NaN. Only the first 10 rows are
    # considered because towards the end of the dataframe, NaN start occuring again.
    num_header_rows = pd.to_numeric(data.iloc[:10, 0], errors="coerce").isna().sum()

    if num_header_rows == 0:
        raise MalformedDataFileError(filename)

    # second pass
    data = pd.read_excel(
        BytesIO(content),
        sheet_name=sheet_name,
        header=list(range(num_header_rows)),  # this creates a MultiIndex header
        index_col=0,
    )

    if data.empty:
        raise MalformedDataFileError(filename)

    # drop comments
    data = data.drop("Time markers", axis=1, errors="ignore")

    # flatten the column names
    data.columns = [
        re.sub(", Unnamed: .*, ", ", ", ", ".join(column).strip().replace("\n", " "))
        for column in data.columns.to_numpy()
    ]

    # The measurements have different numbers of points which causes NaN entries in
    # both measurements and index. Drop all rows where any NaN appears.
    data = data[data.index.notna()].dropna()

    return Path(filename).stem, data


def get_example_dataset_urls() -> Dict[str, bytes]:
    """Get example datasets from package resources.

    Get all example datasets from ``metabolator/static/datasets`` and return them as a
    ``dict`` mapping from filename to data URL.

    If a file's extension is unknown, a ``UnsupportedFileTypeError`` is raised. Files
    whose name contains ``.license`` or ``.gitkeep`` are ignored.
    """
    datasets = {}
    for resource in (
        resources.files("metabolator").joinpath("static").joinpath("datasets").iterdir()
    ):
        if ".license" not in resource.name and ".gitkeep" not in resource.name:
            extension = Path(resource.name).suffix
            if extension not in MIME_TYPES:
                raise UnsupportedFileTypeError(resource.name, extension)
            datasets[resource.name] = data_url(
                MIME_TYPES[extension], resource.read_bytes()
            )

    return datasets
