# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum Dresden - Rossendorf (HZDR)
#
# SPDX-License-Identifier: Apache-2.0

from typing import Tuple

import numpy as np
import numpy.typing as npt
from lmfit import Minimizer, Parameter, Parameters
from lmfit.minimizer import MinimizerResult

# Variance of the range for the four-point method
VARIANCE = 0.45

# Variances of the range of raw data around the first peak
# for the do_fit subroutine
VARIANCE_A = 0.030
VARIANCE_B = 0.030

# Singularity shift in fit function
SHIFT_SINGULARITY = 4

# Maximum number of function evaluations to do in the minimizer
MAX_NUM_FUNC_EVALS = 10000


def model(
    x: npt.NDArray,
    rz: float | Parameter,
    hz: float | Parameter,
    hf: float | Parameter,
    delta: float | Parameter,
) -> npt.NDArray:
    """Model function describing the metabolic rate over time.

    Details are provided in the following publication:
    https://doi.org/10.3390/microorganisms10071397
    """
    return rz * (x + delta) * (hz - (x + delta)) / (hz + hf - (x + delta))


def residual(
    parameters: Parameters, x: npt.NDArray, y: npt.NDArray, weights: npt.NDArray
) -> npt.NDArray:
    """Objective function that is minimized by the do_fit subroutine"""
    rz = parameters["rz"]
    hz = parameters["hz"]
    hf = parameters["hf"]
    delta = parameters["delta"]
    error = model(x, rz, hz, hf, delta) - y
    return error * weights


def get_four_point_optim(
    x: npt.NDArray, y: npt.NDArray
) -> Tuple[npt.NDArray, float, float]:
    """Determine initial fit parameter values using a four point method"""
    varA = []  # noqa: N806
    areaA = []  # noqa: N806
    hoA = []  # noqa: N806
    var = int(len(x) * VARIANCE)

    for i in range(var):
        xr = x[i : -1 - i]
        yr = y[i : -1 - i]

        h1 = xr[0]
        h2 = xr[int(len(xr) * 0.33)]
        h3 = xr[int(len(xr) * 0.66)]
        h4 = xr[-1]

        p1 = yr[0]
        p2 = yr[int(len(xr) * 0.33)]
        p3 = yr[int(len(xr) * 0.66)]
        p4 = yr[-1]

        g1 = p2 * h1 / (p1 * h2)  # noqa: F841
        g2 = p3 * h2 / (p2 * h3)  # noqa: F841
        g3 = p4 * h3 / (p3 * h4)  # noqa: F841

        g = p1 * h4 / (p4 * h1)

        hf1 = []
        hf2 = []
        hf3 = []
        ho = []

        for k in range(len(xr) * 10):
            ho.append(k)
            x1 = k - h1
            x2 = k - h2
            x3 = k - h3
            x4 = k - h4

            hf1.append(x1 * x2 * (g - 1) / (x2 - x1 * g))
            hf2.append(x2 * x3 * (g - 1) / (x3 - x2 * g))
            hf3.append(x3 * x4 * (g - 1) / (x4 - x3 * g))

        idx1 = np.argwhere(np.diff(np.sign(np.array(hf1) - np.array(hf2)))).flatten()
        idx2 = np.argwhere(np.diff(np.sign(np.array(hf2) - np.array(hf3)))).flatten()
        idx3 = np.argwhere(np.diff(np.sign(np.array(hf3) - np.array(hf1)))).flatten()

        if (len(idx1) != 0) & (len(idx2) != 0) & (len(idx3) != 0):
            tx1, ty1 = ho[idx1[0]], hf1[idx1[0]]
            tx2, ty2 = ho[idx2[0]], hf2[idx2[0]]
            tx3, ty3 = ho[idx3[0]], hf3[idx3[0]]

            varA.append(i)
            areaA.append(
                abs(0.5 * (tx1 * (ty2 - ty3) + tx2 * (ty3 - ty1) + tx3 * (ty1 - ty2))),
            )
            hoA.append(ho[idx1[0]])

    if len(areaA) == 0:
        message = "Non-optimal fitting"
        raise RuntimeError(message)

    idx = np.argwhere(areaA == np.min(areaA))
    return idx, hoA[idx.item(0)], g


def do_fit(
    x: npt.NDArray, y: npt.NDArray
) -> Tuple[float, float, MinimizerResult, float]:
    """Main fitting subroutine

    This subroutine fit the model function to the raw data for
    metabolic rate parsed from the excel file.
    The analysis is done by fitting [Monod's growth equation](https://doi.org/10.1146/annurev.mi.03.100149.002103)
    on the data produced by the calorimeter.

    """

    idx, hoA, g = get_four_point_optim(x, y)  # noqa: N806

    tx = np.array(x[idx.item(0) : -1 - idx.item(0)], dtype=float)
    ty = np.array(y[idx.item(0) : -1 - idx.item(0)], dtype=float)

    data_points = np.size(x)
    range_a, range_b = (int(data_points * VARIANCE_A), int(data_points * VARIANCE_B))

    # initialize lmfit
    parameters = Parameters()
    parameters.add(name="rz", value=g, min=0)
    parameters.add(name="hz", value=hoA, min=0)
    parameters.add(name="pole", value=np.max(x) * SHIFT_SINGULARITY, min=np.max(x))

    # Singularity should be shifted out of the fitting range
    # improves speed and stability of the fitting procedure
    parameters.add(name="hf", expr="pole-hz")
    parameters.add(name="delta", value=0.0, vary=True, max=np.inf)

    # Modify the fitting region for best fit according to R-squared (coefficient of
    # determination).
    # See: https://en.wikipedia.org/wiki/Coefficient_of_determination
    # NOTE: Improvement necessary
    best_r_squared = 0
    best_result = None
    best_llim_index = 0
    best_ulim_index = 0
    for llim_index in range(range_a):
        for ulim_index in range(range_b):
            tx = np.array(x[llim_index : -1 - ulim_index], dtype=float)
            ty = np.array(y[llim_index : -1 - ulim_index], dtype=float)

            # Step 1: Determine the number of data points in the selected interval
            num_points = len(tx)

            # Step 2: Calculate dHav (average step size along the enthalpy axis)
            total_enthalpy_interval = tx[-1] - tx[0]
            dHav = total_enthalpy_interval / (num_points - 1)  # noqa: N806

            # Step 3 and 4:
            # Calculate weights based on distance from predecessor and dHav
            dHi = np.diff(tx)  # noqa: N806
            weights = dHi / dHav
            weights = np.insert(weights, 0, 0)  # Insert 0 as weight for the first point

            # Custom weight calculation completed

            minimizer = Minimizer(
                residual,
                parameters,
                fcn_args=(tx, ty, weights),
                max_nfev=MAX_NUM_FUNC_EVALS,
            )
            result = minimizer.leastsq()

            curve_fit = ty + result.residual
            curve_fit[curve_fit > np.max(ty)] = np.max(ty)

            ssr = result.chisqr  # residual sum of squares
            sst = sum(np.power(ty - np.mean(curve_fit), 2))  # total sum of squares
            r_squared = 1 - ssr / sst

            if r_squared > best_r_squared:
                best_r_squared = r_squared
                best_result = result
                best_llim_index = llim_index
                best_ulim_index = ulim_index

    return (
        x[best_llim_index],
        x[-1 - best_ulim_index],
        best_result,
        best_r_squared,
    )
