# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum Dresden - Rossendorf (HZDR)
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

import pandas as pd
import pytest

from metabolator import files


@pytest.fixture
def empty_excel_file():
    filepath = Path(__file__).parent / "empty.xlsx"
    with open(filepath, "br") as f:
        return "empty.xlsx", f.read()


@pytest.fixture
def header_only_excel_file():
    filepath = Path(__file__).parent / "header-only.xlsx"
    with open(filepath, "br") as f:
        return "header-only.xlsx", f.read()


@pytest.fixture
def no_header_excel_file():
    filepath = Path(__file__).parent / "no-header.xlsx"
    with open(filepath, "br") as f:
        return "no-header.xlsx", f.read()


class TestFiles:
    example_file = "chiralLipids_27092023.xls"
    example_file_len = 4475392
    example_file_data_url_len = 5967229

    @pytest.mark.skip("We don't have any example files")
    def test_get_resource_contents(self):
        contents = files.get_resource_contents(self.example_file)
        assert isinstance(contents, bytes)
        assert len(contents) == self.example_file_len

    @pytest.mark.skip("We don't have any example files")
    def test_get_resource_data_url(self):
        url = files.get_resource_data_url(self.example_file)
        assert isinstance(url, str)
        assert url.startswith("data:application/vnd.ms-excel;base64,")
        assert len(url) == self.example_file_data_url_len

    def test_get_resource_data_url_unsupported_file_type(self):
        with pytest.raises(
            files.UnsupportedFileTypeError,
            match=r"Unsupported file type '\.pdf': report\.pdf",
        ):
            _ = files.get_resource_data_url("report.pdf")

    def test_data_url(self):
        url = files.data_url("application/octet-stream", b"hello world\n")
        assert url == "data:application/octet-stream;base64,aGVsbG8gd29ybGQK"

    def test_make_excel_file_data_url(self):
        # Use an empty data frame as this is complicated enough for the data URL to
        # contain all possible characters.
        url = files.make_excel_file_data_url(pd.DataFrame())
        assert url.startswith(
            "data:application/vnd.openxmlformats-officedocument.spreadsheetml.sheet;base64,"
        )
        assert url.count(",") == 1
        # Standard base64 encoding must be used; URL-safe encoding won't work!
        _, encoded_data = url.split(",", maxsplit=1)
        assert "+" in encoded_data and "-" not in encoded_data
        assert "/" in encoded_data and "_" not in encoded_data

    def test_load_excel_file_empty(self, empty_excel_file):
        with pytest.raises(files.MalformedDataFileError):
            _ = files.load_excel_file_data(*empty_excel_file)

    def test_load_excel_file_only_header(self, header_only_excel_file):
        with pytest.raises(files.MalformedDataFileError):
            _ = files.load_excel_file_data(*header_only_excel_file)

    def test_load_excel_file_no_header(self, no_header_excel_file):
        with pytest.raises(files.MalformedDataFileError):
            _ = files.load_excel_file_data(*no_header_excel_file)
